"""This module provides a runner for the AxmpAPIModels."""

import asyncio
import json
import logging
import logging.config

from axmp_openapi_helper.openapi.axmp_api_models import AxmpOpenAPI
from axmp_openapi_helper.utils.converter import Converter

logging.config.fileConfig("logging.conf", disable_existing_loggers=False)
logging.getLogger("axmp_openapi_helper.openapi.multi_openapi_helper").setLevel(
    logging.INFO
)
logging.getLogger("axmp_openapi_helper.openapi.axmp_api_models").setLevel(logging.INFO)
logging.getLogger("axmp_openapi_helper.openapi.operation").setLevel(logging.INFO)
logging.getLogger("axmp_openapi_helper.openapi.fastapi.fastapi_models").setLevel(
    logging.DEBUG
)
logger = logging.getLogger("appLogger")
logger.setLevel(logging.DEBUG)


async def main():
    """Run the axmp api models."""
    # axmp_open_api = AxmpOpenAPI.from_spec_file(
    #     file_path="runner/openapi/zcp/zcp_alert_backend_openapi_spec.json"
    # )

    # openapi is a dict case
    # openapi: dict = json.loads(
    #     open("runner/openapi/zcp/zcp_mcm_backend_openapi_spec.json").read()
    # )
    # axmp_open_api = AxmpOpenAPI.from_spec_dict(data=openapi)

    # testing
    # operations = axmp_open_api.get_operations_by_tag_method_pattern(
    #     tags=["alert", "channel"], methods=["delete"], regex=r"^/api/alert/v1/.*"
    # )
    # test case that causes errors in version 0.1.7
    # operations = axmp_open_api.get_operations_by_tag_method_pattern(
    #     tags=["alert", "channel"], methods=["get","post","put","delete"], regex=r"^/api/v1/csp/{csp}/auth/check"
    # )
    # test case: uppercase methods are supported
    # operations = axmp_open_api.get_operations_by_tag_method_pattern(
    #     tags=["alert", "channel"], methods=["GET"], regex=r"^/aaaaaa"
    # )

    # test case: ignore content tpye application/xml, application/x-www-form-urlencoded
    openapi: dict = json.loads(
        open("runner/openapi/zcp/zcp_alert_backend_openapi_spec.json").read()
    )
    axmp_open_api = AxmpOpenAPI.from_spec_dict(data=openapi)
    operations = axmp_open_api.get_operations_by_tag_method_pattern(
        tags=["alert", "channel"],
        methods=["get", "post", "put", "delete"],
        regex=r"^/aaaaaa",
    )
    tools = Converter.operations_to_tools(
        server_name="zcp_alert_backend",
        base_path="/api/alert/v1",
        axmp_open_api=axmp_open_api,
        operations=operations,
    )

    # print the tools
    for tool in tools:
        print("-" * 100)
        # print(f"server_name: {tool.server_name}")
        print(f"tool_name: {tool.name}")
        # print(f"tool_description: {tool.description}")
        print(f"tool_path: {tool.path}")
        print(f"tool_method: {tool.method}")
        # print(f"tool_query_params: {tool.query_params}")
        # print(f"tool_path_params: {tool.path_params}")
        print(f"tool_request_body: {tool.request_body}")
    print("tool count: ", len(tools))


if __name__ == "__main__":
    asyncio.run(main())



"""This module provides a wrapper for the API."""

import logging
import os
from typing import Any, Dict

from httpx import AsyncClient, BasicAuth, Client, Response
from pydantic import BaseModel

from axmp_openapi_helper.openapi.multi_openapi_spec import AuthenticationType

logger = logging.getLogger(__name__)

_TIMEOUT = os.getenv("HTTPX_DEFAULT_TIMEOUT", 10)


class AxmpAPIWrapper:
    """AxmpAPIWrapper is a wrapper for the API."""

    def __init__(
        self,
        server: str,
        /,
        *,
        headers: dict | None = None,
        cookies: dict | None = None,
        auth_type: AuthenticationType | None = AuthenticationType.NONE,
        username: str | None = None,
        password: str | None = None,
        bearer_token: str | None = None,
        api_key_name: str | None = None,
        api_key_value: str | None = None,
        tls_verify: bool | None = False,
        timeout: int | None = _TIMEOUT,
    ):
        """Initialize the AxmpAPIWrapper.

        Args:
            server (str): The server URL
            headers (dict, optional): The headers for the API request. Defaults to None.
            cookies (dict, optional): The cookies for the API request. Defaults to None.
            auth_type (AuthenticationType, optional): The authentication type. Defaults to AuthenticationType.NONE.
            username (str, optional): The username for the API request. Defaults to None.
            password (str, optional): The password for the API request. Defaults to None.
            bearer_token (str, optional): The bearer token for the API request. Defaults to None.
            api_key_name (str, optional): The api key name for the API request. Defaults to None.
            api_key_value (str, optional): The api key value for the API request. Defaults to None.
            tls_verify (bool, optional): Whether to verify the TLS certificate. Defaults to False.
            timeout (int, optional): The timeout for the API request. Defaults to _TIMEOUT.
        """
        if not server:
            raise ValueError("Server URL is required")

        self._server = server
        self._headers = headers or {}
        self._cookies = cookies or {}
        self._auth_type = auth_type
        self._username = username
        self._password = password
        self._bearer_token = bearer_token
        self._api_key_name = api_key_name
        self._api_key_value = api_key_value
        self._tls_verify = tls_verify
        self._timeout = timeout

        self._auth = None
        self._client = None
        self._async_client = None

        if self._headers.get("Content-Type") is None:
            self._headers.update({"Content-Type": "application/json"})

        if auth_type == AuthenticationType.BASIC:
            if not username or not password:
                raise ValueError(
                    "Username and password are required for Basic authentication"
                )
            self._auth = BasicAuth(username=username, password=password)

        elif auth_type == AuthenticationType.BEARER:
            if not bearer_token:
                raise ValueError("Bearer token is required for Bearer authentication")
            if self._bearer_token is not None:
                self._headers.update({"Authorization": f"Bearer {self._bearer_token}"})

        elif auth_type == AuthenticationType.API_KEY:
            if not api_key_name or not api_key_value:
                raise ValueError(
                    "API key name and value are required for API key authentication"
                )
            self._headers.update({self._api_key_name: self._api_key_value})

    @property
    def async_client(self) -> AsyncClient:
        """Get the async client."""
        if self._async_client is None:
            self._async_client = AsyncClient(
                auth=self._auth,
                base_url=self._server,
                headers=self._headers,
                cookies=self._cookies,
                verify=self._tls_verify,
                timeout=self._timeout,
            )
        return self._async_client

    @property
    def client(self) -> Client:
        """Get the client."""
        if self._client is None:
            self._client = Client(
                auth=self._auth,
                base_url=self._server,
                headers=self._headers,
                cookies=self._cookies,
                verify=self._tls_verify,
                timeout=self._timeout,
            )
        return self._client

    async def close(self) -> None:
        """Close the async and sync clients."""
        if self._async_client is not None:
            await self._async_client.aclose()
            self._async_client = None
        if self._client is not None:
            self._client.close()
            self._client = None

    async def _get_response(self, response: Response) -> Dict[str, Any]:
        """Get response from API.

        Args:
            response (Response): Response from API

        Returns:
            Dict[str, Any]: Response data
        """
        if response.status_code == 200:
            logger.debug(f"Response: {response.json()}")

            return response.json()
        else:
            logger.warning(
                f"Failed to get response: {response.status_code} {response.text}"
            )
            return {
                "result": "failed",
                "code": response.status_code,
                "message": response.text,
            }

    async def run(
        self,
        method: str,
        path: str,
        /,
        *,
        path_params: BaseModel | None = None,
        query_params: BaseModel | None = None,
        request_body: BaseModel | None = None,
    ) -> str:
        """Run the API request asynchronously.

        Args:
            method (str): The HTTP method to use
            path (str): The path to the resource
            path_params (Any, optional): Path parameters for the tool. Defaults to None.
            query_params (Any, optional): Query parameters for the tool. Defaults to None.
            request_body (Any, optional): Request body for the tool. Defaults to None.

        Returns:
            str: Response from the API
        """
        logger.debug(f"Method: {method}, Path: {path}")
        logger.debug("-" * 100)
        logger.debug(f"Path params: {path_params}")
        logger.debug(f"Query params: {query_params}")
        logger.debug(f"Request body: {request_body}")

        if path_params is not None:
            # for path parameters
            # NOTE: exclude_none=True, mode="json" is used to exclude None values and convert to json (e.g. Enum values)
            path = path.format(**path_params.model_dump(exclude_none=True, mode="json"))

            logger.debug(f"Formatted path: {path}")

        logger.debug(
            f"Query params: {query_params.model_dump_json(exclude_none=True) if query_params else None}"
        )
        logger.debug(
            f"Request body: {request_body.model_dump_json(exclude_none=True) if request_body else None}"
        )

        # NOTE: If the request body is an array, we need to get the value for the array
        request_body_array_value = None
        if request_body:
            request_body_array_value = await self._get_request_body_value_for_array(
                request_body
            )

        response = await self.async_client.request(
            method,
            path,
            params=query_params.model_dump(exclude_none=True, mode="json")
            if query_params
            else None,
            json=request_body_array_value
            if request_body_array_value
            else request_body.model_dump(exclude_none=True, mode="json")
            if request_body
            else None,
        )

        response.raise_for_status()

        return await self._get_response(response)

    async def _get_request_body_value_for_array(
        self, request_body: BaseModel
    ) -> list[Any] | None:
        """Get request body value for array.

        Args:
            request_body (BaseModel): Request body

        Returns:
            Any: Request body value
        """
        request_body_value = None
        for i, (field_name, field_info) in enumerate(
            type(request_body).model_fields.items()
        ):
            field_value = getattr(request_body, field_name)
            if i == 0 and field_value is not None:
                if isinstance(field_value, list):
                    request_body_value = [
                        item.model_dump(exclude_none=True, mode="json")
                        if isinstance(item, BaseModel)
                        else item
                        for item in field_value
                    ]
                    break
            else:
                logger.warning(f"Request body value is not array: {field_name}")
                return None

        logger.debug(f"Request body value: {request_body_value}")

        return request_body_value



"""This module provides a helper for working with OpenAPI specifications."""

from .multi_openapi_helper import MultiOpenAPIHelper
from .openapi.multi_openapi_spec import (
    APIMap,
    APIServerConfig,
    AuthConfig,
    AuthenticationType,
    MethodSpec,
    MultiOpenAPISpecConfig,
    RouteMap,
    ToolConfig,
)
from .openapi.operation import AxmpAPIOperation
from .wrapper.api_wrapper import AxmpAPIWrapper

__all__ = [
    "MultiOpenAPIHelper",
    "APIMap",
    "APIServerConfig",
    "AuthConfig",
    "AuthenticationType",
    "MethodSpec",
    "RouteMap",
    "ToolConfig",
    "AxmpAPIOperation",
    "MultiOpenAPISpecConfig",
    "AxmpAPIWrapper",
]



"""This module provides a runner for the ZMP OpenAPI Helper."""

import asyncio
import logging
import logging.config

from axmp_openapi_helper import (
    MultiOpenAPISpecConfig,
)
from axmp_openapi_helper.multi_openapi_helper import MultiOpenAPIHelper
from axmp_openapi_helper.openapi.multi_openapi_spec import (
    AuthConfig,
    AuthenticationType,
)
from axmp_openapi_helper.openapi.operation import AxmpAPIOperation

logging.config.fileConfig("logging.conf", disable_existing_loggers=False)
logging.getLogger("axmp_openapi_helper.openapi.multi_openapi_helper").setLevel(
    logging.INFO
)
logging.getLogger("axmp_openapi_helper.openapi.axmp_api_models").setLevel(logging.INFO)
logging.getLogger("axmp_openapi_helper.openapi.operation").setLevel(logging.INFO)
logging.getLogger("axmp_openapi_helper.openapi.fastapi.fastapi_models").setLevel(
    logging.DEBUG
)
logger = logging.getLogger("appLogger")
logger.setLevel(logging.DEBUG)


async def main():
    """Run the multi openapi helper."""
    multi_openapi_spec_config = MultiOpenAPISpecConfig.from_multi_server_spec_file(
        file_path="runner/openapi/zmp_multi_openapi_spec.json"
    )

    multi_openapi_helper = MultiOpenAPIHelper(
        multi_openapi_spec_config=multi_openapi_spec_config
    )

    api_servers = multi_openapi_helper.openapi_servers
    for server_name, server in api_servers.items():
        print("-" * 100)
        print(server_name)
        print(server.endpoint)
        print(server.base_path)
        print(server.spec_file_path)

    for server_name, server in api_servers.items():
        multi_openapi_helper.update_openapi_server_auth_config(
            server_name=server_name,
            auth_config=AuthConfig(
                type=AuthenticationType.API_KEY,
                api_key_name="X-Access-Key",
                api_key_value="zmp-09bae73d-9f59-491a-ae06-7747e8f79883",
            ),
        )
    multi_openapi_helper.initialize_clients()

    operations: list[AxmpAPIOperation] = multi_openapi_helper.all_operations
    for i, operation in enumerate(operations):
        print("-" * 100)
        print(
            f"{i}:{operation.server_name}:{operation.method}:{operation.name}:{operation.path}"
        )
        # print(operation.description)
        # print(operation.args_schema)
        # print(operation.path_params)
        # print(operation.query_params)
        # print(operation.request_body)

    # response = await multi_openapi_helper.run(
    #     name="get_alerts", args={"priorities": ["P1", "P2"]}
    # )
    # print(response)

    # response = await multi_openapi_helper.run(name="get_clusters", args=None)
    # print(response)

    # tags = multi_openapi_helper.get_all_tags()
    # print("-" * 100)
    # print("all tags")
    # print(tags)

    # tags = multi_openapi_helper.get_tags(server_name="zcp-alert-backend")
    # print("-" * 100)
    # print("tags by server name")
    # print(tags)

    # path_method_operations: list[tuple[str, str, Operation]] = (
    #     multi_openapi_helper.get_all_operations_by_path_pattern(regex="^/api/alert/v1.*")
    # )
    # print("-" * 100)
    # print("operations by path pattern")
    # for path, method, operation in path_method_operations:
    #     print(f"{method}:{path}")

    # path_method_operations: list[tuple[str, str, Operation]] = (
    #     multi_openapi_helper.get_all_operations_by_method(method="put")
    # )
    # print("-" * 100)
    # print("operations by method")
    # for path, method, operation in path_method_operations:
    #     print(f"{method}:{path}")

    # path_method_operations: list[tuple[str, str, Operation]] = (
    #     multi_openapi_helper.get_all_operations_by_tag(tag="alert")
    # )
    # print("-" * 100)
    # print("operations by tag")
    # for path, method, operation in path_method_operations:
    #     print(f"{method}:{path}")

    await multi_openapi_helper.close()


if __name__ == "__main__":
    asyncio.run(main())



"""This module provides a model for the mixed API specification."""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import List

from pydantic import BaseModel


class AuthenticationType(str, Enum):
    """Authentication type model."""

    NONE = "None"
    BASIC = "Basic"
    BEARER = "Bearer"
    API_KEY = "ApiKey"


class MethodSpec(BaseModel):
    """A model for the method specification."""

    method: str
    tool_name: str | None = None
    description: str | None = None


class APIMap(BaseModel):
    """A model for the API simple specification."""

    path: str
    methods: List[str | MethodSpec]


class RouteMap(BaseModel):
    """A model for the route map."""

    methods: List[str] | None = None
    tags: List[str] | None = None
    pattern: str | None = None


class ToolConfig(BaseModel):
    """A model for the tool configuration."""

    api_maps: List[APIMap] | None = None
    route_maps: List[RouteMap] | None = None


class AuthConfig(BaseModel):
    """A model for the authentication configuration."""

    type: AuthenticationType = AuthenticationType.NONE
    username: str | None = None
    password: str | None = None
    api_key_name: str | None = None
    api_key_value: str | None = None
    bearer_token: str | None = None


class APIServerConfig(BaseModel):
    """A model for the API server configuration."""

    server_name: str
    endpoint: str
    base_path: str | None = None
    timeout: float = 10
    tls_verify: bool = True
    spec_file_path: str | Path
    # NOTE: open_api_spec is not needed any more
    # open_api_spec: AxmpOpenAPI | None = None
    tool_config: ToolConfig
    auth_config: AuthConfig


class MultiOpenAPISpecConfig(BaseModel):
    """MultiOpenAPISpecConfig is a model that contains the configuration for the multi-server API specification.

    ```json
    {
        "backends": [
            {
                "server_name": "zcp-alert-backend",
                "endpoint": "https://zcp-alert-backend.com",
                "base_path": "/api/alert/v1",
                "tls_verify": false,
                "timeout": 10,
                "auth_config": {
                    "type": "basic",
                    "username": "admin",
                    "password": "password"
                },
                "spec_file_path": "openapi_spec/zcp_spec/alert_openapi_spec.json",
                "tool_config": {
                    "api_maps": [
                        {
                            "path":"/api/alert/v1/alerts",
                            "methods": ["get", "post"]
                        },
                        {
                            "path":"/api/alert/v1/alerts/webhook",
                            "methods": ["post"]
                        },
                        {
                            "path":"/api/alert/v1/alert/priorities",
                            "methods": ["get"]
                        }
                    ]
                }
            },
            {
                "server_name": "example-backend",
                "endpoint": "https://example-backend.com",
                "base_path": "/api/example/v1",
                "spec_file_path": "openapi_spec/example_spec/test_spec.json",
                "tls_verify": true,
                "timeout": 10,
                "auth_config": {
                    "type": "api_key",
                    "custom_header_api_key_name": "X-API-KEY",
                    "custom_header_api_key_value": "1234567890"
                },
                "tool_config": {
                    "api_maps": [
                        {
                            "path":"/api/example/v1/test",
                            "methods": [
                                {
                                    "method": "post",
                                    "tool_name": "create_test",
                                    "description": "Create test"
                                }
                            ]
                        }
                    ],
                    "route_maps": [
                        {
                            "pattern": r".*",
                            "methods": ["get"],
                            "tags": ["*"]
                        }
                    ]
                }
            },
            {
                "server_name": "zcp-alert-backend",
                "endpoint": "https://zcp-alert-backend.com",
                "base_path": "/api/alert/v1",
                "spec_file_path": "openapi_spec/zcp_spec/alert_openapi_spec.json",
                "tls_verify": true,
                "timeout": 10,
                "auth_config": {
                    "type": "bearer",
                    "bearer_token": "1234567890"
                },
                "tool_config": {
                    "route_maps": [
                        {
                            "pattern": "^/api/alert/v1/channels/.*",
                            "methods": ["get"],
                            "tags": []
                        },
                        {
                            "pattern": "^/api/alert/v1/admin/.*",
                            "methods": ["post"],
                            "tags": ["alerts"]
                        }
                    ]
                }
            }
        ]
    }
    ```
    """

    backends: List[APIServerConfig]

    @classmethod
    def from_multi_server_spec_file(
        cls, file_path: str | Path
    ) -> MultiOpenAPISpecConfig:
        """Load the multi-server API specification from a file."""
        with open(file_path) as f:
            return cls.model_validate_json(f.read())



"""This module provides a helper for the OpenAPI specification."""

from __future__ import annotations

import logging
import re

from axmp_openapi_helper.openapi.axmp_api_models import SUPPORTED_METHODS, AxmpOpenAPI
from axmp_openapi_helper.openapi.fastapi.openapi_models import Operation
from axmp_openapi_helper.openapi.multi_openapi_spec import (
    APIServerConfig,
    AuthConfig,
    AuthenticationType,
    MethodSpec,
    MultiOpenAPISpecConfig,
)
from axmp_openapi_helper.openapi.operation import AxmpAPIOperation
from axmp_openapi_helper.utils.converter import Converter
from axmp_openapi_helper.wrapper.api_wrapper import AxmpAPIWrapper

logger = logging.getLogger(__name__)


class MultiOpenAPIHelper:
    """MultiOpenAPIHelper for ZMP ApiWrapper."""

    def __init__(self, multi_openapi_spec_config: MultiOpenAPISpecConfig):
        """Initialize the OpenAPIHelper."""
        self.multi_openapi_spec_config = multi_openapi_spec_config
        self._validate_multi_openapi_spec_config()

        self._openapi_servers: dict[str, APIServerConfig] = (
            self._initialize_openapi_servers()
        )
        self._all_operations: list[AxmpAPIOperation] = self._initialize_all_operations()
        # NOTE: clients are initialized after initialization
        # self._clients: dict[str, AxmpAPIWrapper] = self._initialize_clients()
        self._clients: dict[str, AxmpAPIWrapper] = {}

    def _validate_multi_openapi_spec_config(self):
        """Validate the multi-server API specification configuration."""
        path_method_set = set()
        for backend in self.multi_openapi_spec_config.backends:
            # check the spec file path and zmp open api
            if not backend.spec_file_path:
                raise ValueError(
                    f"Spec file path is required for backend: {backend.server_name}"
                )

            # check auth config
            if backend.auth_config:
                if backend.auth_config.type not in [
                    AuthenticationType.BASIC,
                    AuthenticationType.BEARER,
                    AuthenticationType.API_KEY,
                    AuthenticationType.NONE,
                ]:
                    raise ValueError(f"Invalid auth type: {backend.auth_config.type}")

                # NOTE: username and password are not required during the initialization because the username and password are set on the mcp server
                # issue https://github.com/cloudz-mp/axmp-openapi-helper/issues/46
                # if backend.auth_config.type == AuthenticationType.BASIC and (
                #     not backend.auth_config.username
                # ):
                #     raise ValueError(
                #         f"Username and password are required for basic auth: {backend.server_name}"
                #     )
                if backend.auth_config.type == AuthenticationType.API_KEY and (
                    not backend.auth_config.api_key_name
                    # NOTE: api_key_value is not required during the initialization because the api_key_value is set on the mcp server
                    # or not backend.auth_config.api_key_value
                ):
                    raise ValueError(
                        f"API key name and value are required for api key auth: {backend.server_name}"
                    )

                # NOTE: bearer token is not required during the initialization because the bearer token is set on the mcp server
                # if (
                #     backend.auth_config.type == AuthenticationType.BEARER
                #     and not backend.auth_config.bearer_token
                # ):
                #     raise ValueError(
                #         f"Bearer token is required for bearer auth: {backend.server_name}"
                #     )

            # check tool config
            if not backend.tool_config:
                raise ValueError(
                    f"Tool config is required for backend: {backend.server_name}"
                )

            if not backend.tool_config.api_maps and not backend.tool_config.route_maps:
                raise ValueError(
                    f"API maps or route maps are required for backend: {backend.server_name}"
                )

            # check api maps
            if backend.tool_config.api_maps:
                for api_map in backend.tool_config.api_maps:
                    # check path
                    if not api_map.path or not api_map.path.startswith("/"):
                        raise ValueError(
                            f"API map path is required and must start with /: {api_map.path}"
                        )
                    if api_map.path.endswith("/"):
                        raise ValueError(
                            f"API map path must not end with /: {api_map.path}"
                        )
                    if backend.base_path and not api_map.path.startswith(
                        backend.base_path
                    ):
                        raise ValueError(
                            f"API map path must start with base path: {api_map.path} and base path is {backend.base_path}"
                        )

                    # check methods
                    for method in api_map.methods:
                        method_name = None
                        if isinstance(method, str):
                            method_name = method.lower()
                            if method_name not in SUPPORTED_METHODS:
                                raise ValueError(f"Invalid method name: {method_name}")
                        elif isinstance(method, MethodSpec):
                            method_name = method.method.lower()
                            if method_name not in SUPPORTED_METHODS:
                                raise ValueError(f"Invalid method name: {method_name}")
                            if not method.tool_name and not method.description:
                                raise ValueError(
                                    f"Tool name and description are required for method: {method_name}"
                                )
                        else:
                            raise ValueError(f"Invalid method type: {type(method)}")

                        path_method = (
                            f"[{backend.server_name}:{api_map.path}:{method_name}]"
                        )
                        if path_method in path_method_set:
                            raise ValueError(
                                f"Duplicate operation found: {path_method}"
                            )
                        path_method_set.add(path_method)

            # check route maps
            if backend.tool_config.route_maps:
                for route_map in backend.tool_config.route_maps:
                    if (
                        not route_map.pattern
                        and not route_map.methods
                        and not route_map.tags
                    ):
                        raise ValueError(
                            f"At least one of pattern, methods, and tags is required for route map: {route_map}"
                        )
                    # check pattern whether it is valid regex
                    if route_map.pattern and route_map.pattern != "":
                        try:
                            re.compile(route_map.pattern)
                        except re.error:
                            raise ValueError(
                                f"Invalid regex pattern: {route_map.pattern}"
                            )

                    # check methods whether it is valid and in supported methods
                    if route_map.methods and len(route_map.methods) > 0:
                        for method in route_map.methods:
                            if method.lower() not in SUPPORTED_METHODS:
                                raise ValueError(f"Invalid method name: {method}")

                    # TODO: check tags whether it is valid and in supported tags

    def _initialize_openapi_servers(self) -> dict[str, APIServerConfig]:
        """Initialize the openapi servers."""
        return {
            backend.server_name: backend
            for backend in self.multi_openapi_spec_config.backends
        }

    def _initialize_all_operations(self) -> list[AxmpAPIOperation]:
        """Initialize the all operations."""
        operations: list[AxmpAPIOperation] = []

        for backend in self.multi_openapi_spec_config.backends:
            axmp_open_api: AxmpOpenAPI = None

            if backend.spec_file_path:
                axmp_open_api = AxmpOpenAPI.from_spec_file(backend.spec_file_path)
            else:
                raise ValueError(
                    f"Spec file path is required for backend: {backend.server_name}"
                )

            # generate operations from api maps
            if backend.tool_config.api_maps:
                operations.extend(
                    self._get_api_operation_from_api_maps(
                        backend=backend, axmp_open_api=axmp_open_api
                    )
                )

            # generate operations from route maps
            if backend.tool_config.route_maps:
                _route_map_operations = self._get_api_operation_from_route_maps(
                    backend=backend, axmp_open_api=axmp_open_api
                )
                # check the duplicate operations by path and method of the _route_map_operations in the operations
                for _operation in _route_map_operations:
                    if _operation.name not in [op.name for op in operations]:
                        operations.append(_operation)

        return operations

    def _get_api_operation_from_route_maps(
        self, *, backend: APIServerConfig, axmp_open_api: AxmpOpenAPI
    ) -> list[AxmpAPIOperation]:
        """Get API operations from route maps."""
        common_operations: list[tuple[str, str, Operation]] = []

        for route_map in backend.tool_config.route_maps:
            pattern_matched_operations: list[tuple[str, str, Operation]] = []
            tag_matched_operations: list[tuple[str, str, Operation]] = []
            method_matched_operations: list[tuple[str, str, Operation]] = []

            if route_map.pattern:
                pattern_matched_operations = (
                    axmp_open_api.get_operations_by_path_pattern(
                        regex=route_map.pattern
                    )
                )

            if route_map.tags and len(route_map.tags) > 0:
                for tag in route_map.tags:
                    tag_matched_operations.extend(
                        axmp_open_api.get_operations_by_tag(tag=tag)
                    )

            if route_map.methods and len(route_map.methods) > 0:
                for method in route_map.methods:
                    method_matched_operations.extend(
                        axmp_open_api.get_operations_by_method(method=method)
                    )

            # extract the common operations from pattern_matched_operations, tag_matched_operations, method_matched_operations
            # should exclude the empty *_matched_operations during the extract the common operations
            if pattern_matched_operations:
                if tag_matched_operations:
                    if method_matched_operations:
                        for path, method, operation in pattern_matched_operations:
                            if (path, method) in [
                                (path, method)
                                for path, method, _ in tag_matched_operations
                            ] and (path, method) in [
                                (path, method)
                                for path, method, _ in method_matched_operations
                            ]:
                                # check the duplicate operations by path and method of the common_operations
                                if (path, method, operation) not in common_operations:
                                    common_operations.append((path, method, operation))
                    else:
                        for path, method, operation in pattern_matched_operations:
                            if (path, method) in [
                                (path, method)
                                for path, method, _ in tag_matched_operations
                            ]:
                                # check the duplicate operations by path and method of the common_operations
                                if (path, method, operation) not in common_operations:
                                    common_operations.append((path, method, operation))
                else:
                    if method_matched_operations:
                        for path, method, operation in pattern_matched_operations:
                            if (path, method) in [
                                (path, method)
                                for path, method, _ in method_matched_operations
                            ]:
                                # check the duplicate operations by path and method of the common_operations
                                if (path, method, operation) not in common_operations:
                                    common_operations.append((path, method, operation))
                    else:
                        for path, method, operation in pattern_matched_operations:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
            else:
                if tag_matched_operations:
                    if method_matched_operations:
                        for path, method, operation in tag_matched_operations:
                            if (path, method) in [
                                (path, method)
                                for path, method, _ in method_matched_operations
                            ]:
                                # check the duplicate operations by path and method of the common_operations
                                if (path, method, operation) not in common_operations:
                                    common_operations.append((path, method, operation))
                    else:
                        for path, method, operation in tag_matched_operations:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
                else:
                    if method_matched_operations:
                        for path, method, operation in method_matched_operations:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
                    else:
                        # NOTE: if all the *_matched_operations are empty, we should raise an error
                        raise ValueError(
                            f"All the *_matched_operations are empty for route map: {route_map}"
                        )

        return self._convert_operations_to_api_operations(
            backend=backend,
            axmp_open_api=axmp_open_api,
            operations=common_operations,
        )

    def _convert_operations_to_api_operations(
        self,
        *,
        backend: APIServerConfig,
        axmp_open_api: AxmpOpenAPI,
        operations: list[tuple[str, str, Operation]],
    ) -> list[AxmpAPIOperation]:
        """Convert operations to API operations."""
        api_operations: list[AxmpAPIOperation] = []
        for path, method, operation in operations:
            tool_name = Converter.generate_name_from_path(
                path=path,
                method=method,
                base_path=backend.base_path,
            )

            description = None
            if operation.description:
                description = operation.description
            else:
                description = ""

            query_params, path_params, request_body = (
                axmp_open_api.generate_models_by_path_and_method(
                    path=path, method=method
                )
            )

            api_operations.append(
                AxmpAPIOperation(
                    server_name=backend.server_name,
                    name=tool_name,
                    description=description,
                    path=path,
                    method=method,
                    query_params=query_params,
                    path_params=path_params,
                    request_body=request_body,
                )
            )

        return api_operations

    def _get_api_operation_from_api_maps(
        self, *, backend: APIServerConfig, axmp_open_api: AxmpOpenAPI
    ) -> list[AxmpAPIOperation]:
        """Get API operations from API maps."""
        operations: list[AxmpAPIOperation] = []

        for api_map in backend.tool_config.api_maps:  # type: ignore
            for method in api_map.methods:
                method_name = None
                tool_name = None
                description = None

                if isinstance(method, str):
                    method_name = method
                elif isinstance(method, MethodSpec):
                    method_name = method.method
                    description = method.description
                    tool_name = method.tool_name
                else:
                    raise ValueError(f"Invalid method type: {type(method)}")

                operation: Operation = axmp_open_api.get_operation_by_path_method(
                    path=api_map.path,
                    method=method_name,
                )
                query_params, path_params, request_body = (
                    axmp_open_api.generate_models_by_path_and_method(
                        path=api_map.path, method=method_name
                    )
                )

                # if tool_name is not provided, generate it from the path and method
                if not tool_name:
                    tool_name = Converter.generate_name_from_path(
                        path=api_map.path,
                        method=method_name,
                        base_path=backend.base_path,
                    )

                # duplicate check the tool_name in the operations
                # because the tool_name should be unique in the operations for the mcp server
                for op in operations:
                    if op.name == tool_name:
                        # NOTE: if the tool_name is duplicate, we should add a number to the tool_name
                        # to make it unique
                        tool_name = self._generate_unique_tool_name(
                            tool_name=tool_name, operations=operations
                        )
                        break

                # if description is not provided, generate it from the operation
                if not description:
                    if operation.description:
                        description = operation.description
                    else:
                        description = ""

                operations.append(
                    AxmpAPIOperation(
                        server_name=backend.server_name,
                        name=tool_name,
                        description=description,
                        path=api_map.path,
                        method=method_name,
                        query_params=query_params,
                        path_params=path_params,
                        request_body=request_body,
                    )
                )

        return operations

    def _generate_unique_tool_name(
        self, *, tool_name: str, operations: list[AxmpAPIOperation]
    ) -> str:
        """Generate the unique tool name."""
        if tool_name in [op.name for op in operations]:
            tool_name_index = tool_name.split("_")[-1]
            if tool_name_index.isdigit():
                tool_name = f"{tool_name.split('_')[:-1]}_{int(tool_name_index) + 1}"
            else:
                tool_name = f"{tool_name}_1"
        return tool_name

    def initialize_clients(self) -> None:
        """Initialize the clients."""
        clients: dict[str, AxmpAPIWrapper] = {}
        for server_name, openapi_server in self._openapi_servers.items():
            if openapi_server.auth_config:
                auth_config = openapi_server.auth_config
                if auth_config.type == AuthenticationType.BASIC:
                    if not auth_config.username:
                        raise ValueError(
                            f"Username is required for basic auth: {server_name}"
                        )
                    if not auth_config.password:
                        raise ValueError(
                            f"Password is required for basic auth: {server_name}"
                        )
                elif auth_config.type == AuthenticationType.API_KEY:
                    if not auth_config.api_key_name:
                        raise ValueError(
                            f"API key name is required for api key auth: {server_name}"
                        )
                    if not auth_config.api_key_value:
                        raise ValueError(
                            f"API key value is required for api key auth: {server_name}"
                        )
                elif auth_config.type == AuthenticationType.BEARER:
                    if not auth_config.bearer_token:
                        raise ValueError(
                            f"Bearer token is required for bearer auth: {server_name}"
                        )

            clients[server_name] = AxmpAPIWrapper(
                openapi_server.endpoint,
                auth_type=openapi_server.auth_config.type,
                username=openapi_server.auth_config.username,
                password=openapi_server.auth_config.password,
                bearer_token=openapi_server.auth_config.bearer_token,
                api_key_name=openapi_server.auth_config.api_key_name,
                api_key_value=openapi_server.auth_config.api_key_value,
                tls_verify=openapi_server.tls_verify,
                timeout=openapi_server.timeout,
            )

        self._clients = clients

    def update_openapi_server_auth_config(
        self, *, server_name: str, auth_config: AuthConfig
    ) -> None:
        """Update the auth config."""
        self._openapi_servers[server_name].auth_config = auth_config

    @property
    def openapi_servers(self) -> dict[str, APIServerConfig]:
        """Get the openapi servers."""
        return self._openapi_servers

    @property
    def all_operations(self) -> list[AxmpAPIOperation]:
        """Generate the operations from the multi-server API specification configuration."""
        if not self._all_operations:
            self._initialize_all_operations()

        return self._all_operations

    @property
    def clients(self) -> dict[str, AxmpAPIWrapper]:
        """Get the clients."""
        return self._clients

    def get_operations_by_server_name(
        self, *, server_name: str
    ) -> list[AxmpAPIOperation]:
        """Get the operations by server name."""
        return [op for op in self.all_operations if op.server_name == server_name]

    # NOTE: @deprecated, use Converter.generate_name_from_path instead
    def _generate_name_from_path(
        self, *, path: str, method: str, base_path: str | None = None
    ) -> str:
        """Generate the operation name from the path."""
        if base_path:
            if path.startswith(base_path):
                path = path.replace(base_path, "")
            else:
                # NOTE: if the path does not start with the base_path, it means the path is not in the base_path
                # e.g. /healthz is not in the base_path /api/alert/v1
                # raise ValueError(f"Path {path} does not start with prefix {base_path}")
                pass

        replaced_path = re.sub(r"[{}]", "", path)  # remove path params brackets
        replaced_path = re.sub(r"[/:~$-]", "_", replaced_path)  # replace /,:,$,~ with _

        return f"{method.lower()}{replaced_path}"

    async def run(self, *, name: str, args: dict | None = None) -> str:
        """Run the operation by name and args."""
        logger.debug(f"name: {name}")
        logger.debug(f"args: {args}")

        operation = next((op for op in self.all_operations if op.name == name), None)
        if not operation:
            raise ValueError(f"Operation {name} not found")

        if args is None:
            args = {}

        path_params = operation.path_params(**args) if operation.path_params else None
        query_params = (
            operation.query_params(**args) if operation.query_params else None
        )
        request_body = (
            operation.request_body(**args) if operation.request_body else None
        )

        logger.debug(f"path_params: {path_params}")
        logger.debug(f"query_params: {query_params}")
        logger.debug(f"request_body: {request_body}")

        client = self._clients[operation.server_name]

        logger.debug(f"client: {client}")

        return await client.run(
            operation.method,
            operation.path,
            path_params=path_params,
            query_params=query_params,
            request_body=request_body,
        )

    def get_all_tags(self) -> list[str]:
        """Get all tags of the operations."""
        tags = []

        for backend in self.multi_openapi_spec_config.backends:
            axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
                backend.spec_file_path
            )

            tags.extend(axmp_open_api.get_tags())

        return list(set(tags))

    def get_tags(self, *, server_name: str) -> list[str]:
        """Get all tags of the operations by server name."""
        open_api_server = self._openapi_servers[server_name]
        axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
            open_api_server.spec_file_path
        )

        return axmp_open_api.get_tags()

    # get all operations by tag
    def get_all_operations_by_tag(
        self, *, tag: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by tag."""
        operations: list[tuple[str, str, Operation]] = []
        for backend in self.multi_openapi_spec_config.backends:
            axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
                backend.spec_file_path
            )

            operations.extend(axmp_open_api.get_operations_by_tag(tag=tag))

        return operations

    def get_operations_by_tag(
        self, *, server_name: str, tag: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by tag."""
        open_api_server = self._openapi_servers[server_name]
        axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
            open_api_server.spec_file_path
        )

        return axmp_open_api.get_operations_by_tag(tag=tag)

    def get_all_operations_by_path_pattern(
        self, *, regex: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by path pattern."""
        operations: list[tuple[str, str, Operation]] = []
        for backend in self.multi_openapi_spec_config.backends:
            axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
                backend.spec_file_path
            )

            operations.extend(axmp_open_api.get_operations_by_path_pattern(regex=regex))

        return operations

    def get_operations_by_path_pattern(
        self, *, server_name: str, regex: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by path pattern."""
        open_api_server = self._openapi_servers[server_name]
        axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
            open_api_server.spec_file_path
        )

        return axmp_open_api.get_operations_by_path_pattern(regex=regex)

    def get_all_operations_by_method(
        self, *, method: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by method."""
        operations: list[tuple[str, str, Operation]] = []
        for backend in self.multi_openapi_spec_config.backends:
            axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
                backend.spec_file_path
            )

            operations.extend(axmp_open_api.get_operations_by_method(method=method))

        return operations

    def get_operations_by_method(
        self, *, server_name: str, method: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by method."""
        open_api_server = self._openapi_servers[server_name]
        axmp_open_api: AxmpOpenAPI = AxmpOpenAPI.from_spec_file(
            open_api_server.spec_file_path
        )

        return axmp_open_api.get_operations_by_method(method=method)

    async def close(self) -> None:
        """Close the clients."""
        for client in self._clients.values():
            await client.close()



"""Converter for OpenAPI operations to API operations."""

import re

from axmp_openapi_helper.openapi.axmp_api_models import AxmpOpenAPI
from axmp_openapi_helper.openapi.fastapi.openapi_models import Operation
from axmp_openapi_helper.openapi.operation import AxmpAPIOperation


class Converter:
    """Converter for OpenAPI operations to API operations."""

    @classmethod
    def operations_to_tools(
        cls,
        *,
        server_name: str,
        base_path: str,
        axmp_open_api: AxmpOpenAPI,
        operations: list[tuple[str, str, Operation]],
    ) -> list[AxmpAPIOperation]:
        """Convert operations to API operations."""
        tools: list[AxmpAPIOperation] = []
        for path, method, operation in operations:
            tool_name = cls.generate_name_from_path(
                path=path,
                method=method,
                base_path=base_path,
            )

            description = None
            if operation.description:
                description = operation.description
            else:
                description = ""

            query_params, path_params, request_body = (
                axmp_open_api.generate_models_by_path_and_method(
                    path=path, method=method
                )
            )

            tools.append(
                AxmpAPIOperation(
                    server_name=server_name,
                    name=tool_name,
                    description=description,
                    path=path,
                    method=method,
                    query_params=query_params,
                    path_params=path_params,
                    request_body=request_body,
                )
            )

        return tools

    @classmethod
    def generate_name_from_path(
        cls, *, path: str, method: str, base_path: str | None = None
    ) -> str:
        """Generate the operation name from the path."""
        if base_path and base_path != "/":
            if path.startswith(base_path):
                path = path.replace(base_path, "")
            else:
                # NOTE: if the path does not start with the base_path, it means the path is not in the base_path
                # e.g. /healthz is not in the base_path /api/alert/v1
                # raise ValueError(f"Path {path} does not start with prefix {base_path}")
                pass

        replaced_path = re.sub(r"[{}]", "", path)  # remove path params brackets
        replaced_path = re.sub(r"[/:~$-]", "_", replaced_path)  # replace /,:,$,~ with _

        if method.lower() == "get":
            return f"get{replaced_path}"
        elif method.lower() == "post":
            return f"create{replaced_path}"
        elif method.lower() == "put":
            return f"update{replaced_path}"
        elif method.lower() == "delete":
            return f"delete{replaced_path}"
        elif method.lower() == "patch":
            return f"patch{replaced_path}"
        else:
            return f"{method.lower()}{replaced_path}"



"""This module provides a model for the OpenAPI specification."""

import json
import logging
import re
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Dict, List, Tuple, Type, Union

from pydantic import BaseModel, ConfigDict, Field, create_model

from axmp_openapi_helper.openapi.fastapi.openapi_models import (
    OpenAPI,
    Operation,
    Parameter,
    ParameterInType,
    PathItem,
    Reference,
    Schema,
    SchemaOrBool,
)

logger = logging.getLogger(__name__)

SUPPORTED_METHODS: list[str] = [
    "get",
    "put",
    "post",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
]


class SchemaType(Enum):
    """SchemaType is an enum for the schema type."""

    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    OBJECT = "object"
    ARRAY = "array"
    NULL = "null"


class AxmpOpenAPI(OpenAPI):
    """AxmpOpenAPI is a subclass of OpenAPI."""

    @classmethod
    def from_spec_file(cls, file_path: str | Path):
        """Load OpenAPI from file.

        Args:
            file_path (str | Path): File path.

        Returns:
            AxmpOpenAPI: AxmpOpenAPI instance.
        """
        with open(file_path) as file:
            openapi = cls.model_validate_json(file.read())

            # convert paths to PathItem enforced
            if openapi.paths:
                converted_paths = {}
                for path, path_item in openapi.paths.items():
                    if isinstance(path_item, dict):
                        converted_paths[path] = PathItem.model_validate(path_item)
                    else:
                        converted_paths[path] = path_item
                openapi.paths.update(converted_paths)

            return openapi

    @classmethod
    def from_spec_string(cls, spec_string: str):
        """Load OpenAPI from specification string.

        Args:
            spec_string (str): OpenAPI specification string.

        Returns:
            AxmpOpenAPI: AxmpOpenAPI instance.
        """
        openapi = cls.model_validate_json(spec_string)
        # convert paths to PathItem enforced
        if openapi.paths:
            converted_paths = {}
            for path, path_item in openapi.paths.items():
                if isinstance(path_item, dict):
                    converted_paths[path] = PathItem.model_validate(path_item)
                else:
                    converted_paths[path] = path_item
            openapi.paths.update(converted_paths)

        return openapi

    @classmethod
    def from_spec_dict(cls, data: dict):
        """Load OpenAPI from specification dictionary.

        Args:
            spec_dict (dict): OpenAPI specification dictionary.
        """
        openapi = cls.model_validate_json(json.dumps(data, ensure_ascii=False))
        # convert paths to PathItem enforced
        if openapi.paths:
            converted_paths = {}
            for path, path_item in openapi.paths.items():
                if isinstance(path_item, dict):
                    converted_paths[path] = PathItem.model_validate(path_item)
                else:
                    converted_paths[path] = path_item
            openapi.paths.update(converted_paths)

        return openapi

    # NOTE: @deprecated, use from_spec_file or from_spec_string instead
    @classmethod
    def from_openapi(cls, openapi: OpenAPI):
        """Load OpenAPI from OpenAPI.

        Args:
            openapi (OpenAPI): OpenAPI.

        Returns:
            AxmpOpenAPI: AxmpOpenAPI instance.
        """
        openapi = cls.model_validate_json(openapi.model_dump_json())  # type: ignore
        # convert paths to PathItem enforced
        if openapi.paths:
            converted_paths = {}
            for path, path_item in openapi.paths.items():
                if isinstance(path_item, dict):
                    converted_paths[path] = PathItem.model_validate(path_item)
                else:
                    converted_paths[path] = path_item
            openapi.paths.update(converted_paths)

        return openapi

    def generate_models_by_path_and_method(
        self,
        *,
        path: str,
        method: str,
    ) -> Tuple[BaseModel | None, BaseModel | None, BaseModel | None]:
        """Generate parameter and request body models from operation and component schema.

        Returns:
            Tuple[BaseModel, BaseModel, BaseModel]: Tuple of parameter and request body models.
                First element is query parameter model.
                Second element is path parameter model.
                Third element is request body model.
        """
        operation = self.get_operation_by_path_method(path=path, method=method)

        query_parameter_model = None
        path_parameter_model = None
        request_body_model = None

        query_parameters = []
        path_parameters = []

        if operation.parameters:
            for parameter in operation.parameters:
                if parameter.in_ is ParameterInType.query:
                    query_parameters.append(parameter)
                elif parameter.in_ is ParameterInType.path:
                    path_parameters.append(parameter)

        if len(query_parameters) > 0:
            query_parameter_model = self._generate_model_from_parameters(
                path=path,
                method=method,
                parameters=query_parameters,
                parameter_in_type=ParameterInType.query,
            )
        else:
            query_parameter_model = None

        if len(path_parameters) > 0:
            path_parameter_model = self._generate_model_from_parameters(
                path=path,
                method=method,
                parameters=path_parameters,
                parameter_in_type=ParameterInType.path,
            )
        else:
            path_parameter_model = None

        if operation.requestBody:
            # Handle both RequestBody and Reference types
            request_body = operation.requestBody
            
            # If requestBody is a Reference, resolve it
            if hasattr(request_body, 'ref') and request_body.ref:
                # This is a Reference type, need to resolve it
                ref = request_body.ref
                class_name, component = self._get_component_schema_by_ref(ref=ref)
                if component and hasattr(component, 'content'):
                    # The component is a RequestBody with content
                    content_ = component.content
                else:
                    logger.warning(f"Reference not found or invalid: {ref}")
                    request_body_model = None
            else:
                # This is a direct RequestBody type
                content_ = request_body.content
            
            if content_:
                schema_ = None
                # Process only the first type in the content array and ignore subsequent types.
                for i, (_, media_type) in enumerate(content_.items()):
                    if i == 0:
                        schema_ = media_type.schema_
                        break
                if schema_:
                    if schema_.title is None:
                        schema_.title = path+"-"+method
                    request_body_model = self._generate_model_from_request_body_schema(
                        schema_=schema_
                    )
                else:
                    logger.warning("Schema is not found for content type")
                    request_body_model = None
            else:
                request_body_model = None
        else:
            request_body_model = None

        return query_parameter_model, path_parameter_model, request_body_model

    def get_operation_by_path_method(self, *, path: str, method: str) -> Operation:
        """Get operation from path.

        Args:
            path (str): Path.

        Returns:
            Operation: Operation.
        """
        method = method.lower()

        logger.debug(f"path: {path}, method: {method}")
        path_item = self.paths[path]

        if path_item is None:
            raise ValueError(f"Path {path} not found")

        logger.debug(
            f"path_item ::\n{path}: {path_item.model_dump_json(indent=4, exclude_none=True)}"
        )

        operation = None
        if isinstance(path_item, PathItem):
            if hasattr(path_item, method):
                operation = getattr(path_item, method)
                if operation is None:
                    raise ValueError(f"Method {method} not found in path {path}")
            else:
                raise ValueError(f"Method {method} not found in path {path}")
        else:
            raise ValueError(f"Path item {path_item} is not a PathItem")

        logger.debug(
            f"operation ::\n{method}: {operation.model_dump_json(indent=4, exclude_none=True)}"
        )

        return operation

    # get all tags of the operation of the pathitem of the paths of openapi spec
    def get_tags(self) -> list[str]:
        """Get all tags of the operation of the pathitem of the paths of openapi spec."""
        tags = []
        for path, path_item in self.paths.items():
            for method in SUPPORTED_METHODS:
                if hasattr(path_item, method):
                    operation: Operation = getattr(path_item, method)  # type: ignore
                    if operation:
                        if operation.tags:
                            tags.extend(operation.tags)

        return list(set(tags))

    # get operations by tag
    def get_operations_by_tag(self, *, tag: str) -> list[tuple[str, str, Operation]]:
        """Get operations by tag.

        Args:
            tag (str): Tag.

        Returns:
            list[tuple[str, str, Operation]]: List of operations.
                First element is path.
                Second element is method.
                Third element is operation.
        """
        logger.debug(f"tag: {tag}")
        if not tag:
            return []

        operations: list[tuple[str, str, Operation]] = []
        for path, path_item in self.paths.items():
            for method in SUPPORTED_METHODS:
                if hasattr(path_item, method):
                    operation: Operation = getattr(path_item, method)  # type: ignore
                    if operation:
                        if operation.tags:
                            if tag in operation.tags:
                                operations.append((path, method, operation))  # type: ignore
        return operations

    # get operations by path regular expression
    def get_operations_by_path_pattern(
        self, *, regex: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by path regular expression.

        Args:
            regex (str): Regular expression.

        Returns:
            list[tuple[str, str, Operation]]: List of operations.
                First element is path.
                Second element is method.
                Third element is operation.
        """
        logger.debug(f"regex: {regex}")
        if not regex:
            return []

        operations: list[tuple[str, str, Operation]] = []
        for path, path_item in self.paths.items():
            if re.match(regex, path):
                for method in SUPPORTED_METHODS:
                    if hasattr(path_item, method):
                        operation: Operation = getattr(path_item, method)  # type: ignore
                        if operation:
                            operations.append((path, method, operation))  # type: ignore
        return operations

    # get operations by method
    def get_operations_by_method(
        self, *, method: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by method.

        Args:
            method (str): Method.

        Returns:
            list[tuple[str, str, Operation]]: List of operations.
                First element is path.
                Second element is method.
                Third element is operation.
        """
        logger.debug(f"method: {method}")
        if not method:
            return []
        
        # Convert method to lowercase if it's uppercase
        method = method.lower()
        
        operations: list[tuple[str, str, Operation]] = []
        for path, path_item in self.paths.items():
            if hasattr(path_item, method):
                operation: Operation = getattr(path_item, method)  # type: ignore
                if operation:
                    operations.append((path, method, operation))  # type: ignore
        return operations

    def get_operations_by_tag_method_pattern(
        self, *, tags: list[str], methods: list[str], regex: str
    ) -> list[tuple[str, str, Operation]]:
        """Get operations by path and method.

        Args:
            tags (list[str]): Tags.
            methods (list[str]): Methods.
            regex (str): Regular expression.

        Returns:
            list[tuple[str, str, Operation]]: List of operations.
                First element is path.
                Second element is method.
                Third element is operation.
        """
        logger.debug(f"tags: {tags}, methods: {methods}, regex: {regex}")
        common_operations: list[tuple[str, str, Operation]] = []

        tag_matched_operations: list[tuple[str, str, Operation]] = []
        if tags:
            for tag in tags:
                tag_matched_operations.extend(self.get_operations_by_tag(tag=tag))

        method_matched_operations: list[tuple[str, str, Operation]] = []
        if methods:
            for method in methods:
                method_matched_operations.extend(
                    self.get_operations_by_method(method=method)
                )

        pattern_matched_operations: list[tuple[str, str, Operation]] = (
            self.get_operations_by_path_pattern(regex=regex)
        )

        if pattern_matched_operations and len(pattern_matched_operations) > 0:
            if tag_matched_operations and len(tag_matched_operations) > 0:
                if method_matched_operations and len(method_matched_operations) > 0:
                    for path, method, operation in pattern_matched_operations:
                        if (path, method) in [
                            (path, method) for path, method, _ in tag_matched_operations
                        ] and (path, method) in [
                            (path, method)
                            for path, method, _ in method_matched_operations
                        ]:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
                else:
                    for path, method, operation in pattern_matched_operations:
                        if (path, method) in [
                            (path, method) for path, method, _ in tag_matched_operations
                        ]:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
            else:
                if method_matched_operations and len(method_matched_operations) > 0:
                    for path, method, operation in pattern_matched_operations:
                        if (path, method) in [
                            (path, method)
                            for path, method, _ in method_matched_operations
                        ]:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
                else:
                    for path, method, operation in pattern_matched_operations:
                        # check the duplicate operations by path and method of the common_operations
                        if (path, method, operation) not in common_operations:
                            common_operations.append((path, method, operation))
        else:
            if tag_matched_operations and len(tag_matched_operations) > 0:
                if method_matched_operations and len(method_matched_operations) > 0:
                    for path, method, operation in tag_matched_operations:
                        if (path, method) in [
                            (path, method)
                            for path, method, _ in method_matched_operations
                        ]:
                            # check the duplicate operations by path and method of the common_operations
                            if (path, method, operation) not in common_operations:
                                common_operations.append((path, method, operation))
                else:
                    for path, method, operation in tag_matched_operations:
                        # check the duplicate operations by path and method of the common_operations
                        if (path, method, operation) not in common_operations:
                            common_operations.append((path, method, operation))
            else:
                if method_matched_operations and len(method_matched_operations) > 0:
                    for path, method, operation in method_matched_operations:
                        # check the duplicate operations by path and method of the common_operations
                        if (path, method, operation) not in common_operations:
                            common_operations.append((path, method, operation))
                else:
                    # NOTE: if all the *_matched_operations are empty, we should raise an error
                    logger.warning(
                        f"All the *_matched_operations are empty for tags: {tags}, methods: {methods}, regex: {regex}"
                    )
                    return []

        return common_operations

    def _generate_model_from_request_body_schema(
        self, *, schema_: Schema
    ) -> BaseModel | None:
        """Generate model from request body schema.

        Args:
            schema_ (Schema): Schema.

        Returns:
            Optional[BaseModel]: Model.
        """
        model_ = None
        if schema_.type:
            if schema_.type == SchemaType.ARRAY.value:
                ref = schema_.items.ref
                if ref:
                    field_type = self._get_field_type_from_schema(schema_=schema_)

                    # @TODO: enhance field definition
                    fields = {}
                    fields[schema_.title.lower()] = Annotated[
                        field_type, Field(default=...)
                    ]

                    logger.info(
                        f"Model for request body(array): {schema_.title} will be created"
                        f". field_type: {field_type}"
                    )

                    model_ = create_model(
                        schema_.title,
                        **fields,
                        __config__=ConfigDict(use_enum_values=True),
                    )
        else:
            ref = schema_.ref
            if ref:
                class_name, component_schema = self._get_component_schema_by_ref(
                    ref=ref
                )
                model_ = self._generate_model_from_component_schema(
                    class_name=class_name, schema_=component_schema
                )
            else:
                logger.warning(f"Ref is not found: {schema_.ref}")
                raise ValueError(f"Ref is not found: {schema_.ref}")

        return model_

    def _generate_model_from_parameters(
        self,
        *,
        path: str,
        method: str,
        parameter_in_type: ParameterInType,
        parameters: List[Parameter | Reference] | None,
    ) -> Type[BaseModel] | None:
        """Generate model from parameters.

        Args:
            operation_id (str): Operation ID.
            parameter_in_type (ParameterInType): Parameter in type.
                Can be query, path, header, cookie.
            parameters (Optional[List[Union[Parameter, Reference]]]): Parameters.

        Returns:
            Type[BaseModel]: Generated model.
                If parameter_in_type is query, the model is query parameter model.
                If parameter_in_type is path, the model is path parameter model.
        """
        if parameters is None:
            return None

        if len(parameters) == 0:
            return None

        fields = {}
        for parameter in parameters:
            logger.debug(
                f"parameter: {parameter.model_dump_json(indent=4, exclude_none=True)}"
            )

            if isinstance(parameter, Reference):
                ...  # don't support reference
            else:
                field_name = parameter.name
                # @TODO: enhance field definition
                field_type, field = self._make_field_from_parameter(parameter=parameter)
                fields[field_name] = Annotated[field_type, field]

                logger.debug(f"field_name: {field_name}")
                logger.debug(f"field: {fields[field_name]}")

        replaced_path = re.sub(r"[/{}_]", "-", path)  # remove path params brackets
        model_name = "".join(word.capitalize() for word in replaced_path.split("-"))
        model_name = f"{model_name}{parameter_in_type.value.capitalize()}Parameters"

        log_str = (
            f"Model for {parameter_in_type.value} parameters: {model_name} will be created\n"
            f"Fields: {len(fields)}\n"
        )
        for field_name, field_info in fields.items():
            log_str += f"  - {field_name}: {field_info}\n"

        logger.info(log_str)

        return create_model(
            model_name, **fields, __config__=ConfigDict(use_enum_values=True)
        )

    def _make_field_from_parameter(
        self,
        *,
        parameter: Parameter,
    ) -> Tuple[Type[Any], Field]:
        """Make field from parameter.

        Args:
            parameter_in_type (ParameterInType): Parameter in type.
            parameter (Parameter): Parameter.

        Returns:
            Tuple[Any, Field]: Tuple of field type and field.
        """
        logger.debug(f"parameter: {parameter}")

        # This schema is the schema of the parameter
        field_schema: Schema = parameter.schema_

        # NOTE: append parameter description and name into the field description and title
        # because the enum type description is retouched by the _get_field_type_from_schema
        # so we need to append it before call _get_field_type_from_schema
        # Issue: https://github.com/cloudz-mp/zmp-zcp-toolkit/issues/11
        if (field_schema.description is None) and (parameter.description is not None):
            field_schema.description = parameter.description
        if (field_schema.title is None) and (parameter.name is not None):
            field_schema.title = parameter.name

        field_type = self._get_field_type_from_schema(schema_=field_schema)

        field = self._make_field_from_schema(
            schema_=field_schema, required=parameter.required
        )

        logger.debug(f"field_type: {field_type}")
        logger.debug(f"field: {field}")

        return (field_type, field)

    def _get_field_type_from_schema(
        self,
        *,
        schema_: Schema,
    ) -> Type[Any]:
        """Get field type from schema.

        Args:
            schema_ (Schema): Schema.

        Returns:
            Type[Any]: Field type.
        """
        logger.debug(f"schema_.type: {schema_.type}")

        field_type = None
        if schema_.type == SchemaType.STRING.value:
            field_type = str
        elif schema_.type == SchemaType.NUMBER.value:
            # NOTE: llm needs float for int when validate because of the max_digit
            field_type = Decimal
        elif schema_.type == SchemaType.INTEGER.value:
            # NOTE: llm needs float for int when validate because of the multipleOf,
            # maximum, minimum, exclusiveMaximum, exclusiveMinimum instead of int
            field_type = Decimal
        elif schema_.type == SchemaType.BOOLEAN.value:
            field_type = bool
        elif schema_.type == SchemaType.OBJECT.value:
            if schema_.additionalProperties:
                if isinstance(schema_.additionalProperties, Schema):
                    if schema_.additionalProperties.type:
                        field_type = dict[
                            str,
                            str
                            if schema_.additionalProperties.type
                            == SchemaType.STRING.value
                            else schema_.additionalProperties.type,
                        ]
                    else:
                        field_type = dict
                elif isinstance(schema_.additionalProperties, dict):
                    field_type = dict[
                        str,
                        str
                        if schema_.additionalProperties["type"]
                        == SchemaType.STRING.value
                        else schema_.additionalProperties["type"],
                    ]
                else:
                    field_type = dict
            else:
                field_type = dict
        elif schema_.type == SchemaType.ARRAY.value:
            if schema_.items:
                if isinstance(schema_.items, Schema):
                    if schema_.items.type:
                        if schema_.items.type == SchemaType.STRING.value:
                            field_type = list[str]
                        elif schema_.items.type == SchemaType.INTEGER.value:
                            # NOTE: llm needs float for int when validate because of the multipleOf,
                            # maximum, minimum, exclusiveMaximum, exclusiveMinimum instead of int
                            field_type = list[Decimal]
                        elif schema_.items.type == SchemaType.NUMBER.value:
                            # NOTE: llm needs float for int when validate because of the max_digit
                            field_type = list[Decimal]
                        elif schema_.items.type == SchemaType.BOOLEAN.value:
                            field_type = list[bool]
                        elif schema_.items.type == SchemaType.OBJECT.value:
                            # TODO: check if the additionalProperties is exist or not
                            field_type = list[dict]
                        else:
                            raise ValueError(
                                f"Unsupported array item type: {schema_.items.type}"
                            )
                    else:
                        if schema_.items.ref:
                            class_name, component_schema = (
                                self._get_component_schema_by_ref(ref=schema_.items.ref)
                            )
                            model_ = self._generate_model_from_component_schema(
                                class_name=class_name, schema_=component_schema
                            )

                            logger.debug(f"model_: {model_}")
                            logger.debug(f"type(model_): {type(model_)}")

                            field_type = list[model_]

                            # Append enum items into the description for LLM
                            # TODO: refactor this in another function
                            if isinstance(model_, type(Enum)):
                                schema_.description = (
                                    f"{schema_.description + '.' if schema_.description else ''}"
                                    f"Values are {', '.join(v.value for v in model_)}"
                                )
                        elif schema_.items.oneOf:
                            one_of = schema_.items.oneOf
                            types = []
                            for item in one_of:
                                if item.ref:
                                    class_name, component_schema = (
                                        self._get_component_schema_by_ref(ref=item.ref)
                                    )
                                    model_ = self._generate_model_from_component_schema(
                                        class_name=class_name, schema_=component_schema
                                    )
                                    types.append(model_)
                                else:
                                    logger.warning(
                                        f"Unsupported array item type because of ref is not found: {item}"
                                    )
                                    raise ValueError(
                                        f"Unsupported array item type because of ref is not found: {item}"
                                    )
                            if len(types) == 0:
                                field_type = Any
                            elif len(types) == 1:
                                field_type = list[types[0]]
                            else:
                                field_type = list[Union[tuple(types)]]
                        else:
                            logger.warning(
                                f"Unsupported array type because of ref is not found: {schema_.items}"
                            )
                            field_type = Any
                elif isinstance(schema_.items, dict):
                    field_type = list[dict]
            else:
                logger.warning(f"Type is array but items is not found: {schema_.items}")
                field_type = Any
        else:
            field_type = self._get_field_type_from_none_type_schema(schema_=schema_)

        return field_type

    def _get_field_type_from_none_type_schema(
        self,
        *,
        schema_: Schema,
    ) -> Type[Any]:
        """Get field type from none schema.

        Args:
            schema_ (Schema): Schema.

        Returns:
            Type[Any]: Field type.
        """
        if schema_.type is not None:
            if schema_.type != "null":
                # raise ValueError(f"Schema has type: {schema_.type}")
                logger.warning(f"Schema has type: {schema_.type}")

        all_of: List[SchemaOrBool] = schema_.allOf
        one_of: List[SchemaOrBool] = schema_.oneOf
        any_of: List[SchemaOrBool] = schema_.anyOf
        ref: str = schema_.ref

        if any_of:
            types = []
            for f_schema in any_of:
                # This is the schema of the parameter's schema (anyOf)
                if isinstance(f_schema, Schema):
                    f_type = self._get_field_type_from_schema(schema_=f_schema)
                    types.append(f_type)

                    # Extract other properties from f_schema and send to parent schema
                    # format, maxLength, minLength, pattern, etc.
                    for field_name, field_value in f_schema.model_dump(
                        exclude_none=True
                    ).items():
                        setattr(schema_, field_name, field_value)

            if len(types) == 0:
                field_type = Any
            elif len(types) == 1:
                field_type = types[0]
            else:
                field_type = Union[tuple(types)]
        elif one_of:
            # TODO: support oneOf, check using the real example
            ...
        elif all_of:
            types = []
            for f_schema in all_of:
                # This is the schema of the parameter's schema (allOf)
                if isinstance(f_schema, Schema):
                    if f_schema.type is not None:
                        raise ValueError(f"Schema has type: {f_schema.type}")
                    else:
                        class_name, component_schema = (
                            self._get_component_schema_by_ref(ref=f_schema.ref)
                        )
                        model_ = self._generate_model_from_component_schema(
                            class_name=class_name, schema_=component_schema
                        )

                        logger.debug(f"model_: {model_}")
                        logger.debug(f"type(model_): {type(model_)}")

                        types.append(model_)

                        # Append enum items into the description for LLM
                        if isinstance(model_, type(Enum)):
                            schema_.description = (
                                f"{schema_.description + '.' if schema_.description else ''}"
                                f"Values are {', '.join(v.value for v in model_)}"
                            )
                else:
                    raise ValueError(f"Unsupported schema type: {type(f_schema)}")

            if len(types) == 0:
                field_type = Any
            elif len(types) == 1:
                field_type = types[0]
            else:
                field_type = Union[tuple(types)]
        elif ref:
            class_name, component_schema = self._get_component_schema_by_ref(ref=ref)
            model_ = self._generate_model_from_component_schema(
                class_name=class_name, schema_=component_schema
            )

            logger.debug(f"model_: {model_}")
            logger.debug(f"type(model_): {type(model_)}")

            # Append enum items into the description for LLM
            if isinstance(model_, type(Enum)):
                schema_.description = (
                    f"{schema_.description + '.' if schema_.description else ''}"
                    f"Values are {', '.join(v.value for v in model_)}"
                )

            field_type = model_
        else:
            field_type = None

        return field_type

    def _make_field_from_schema(
        self,
        *,
        schema_: Schema,
        required: bool,
    ) -> Field:
        """Make field from schema.

        Args:
            schema_ (Schema): Schema.
            required (bool): Required.

        Returns:
            Field: Field of pydantic.
        """
        field = Field(
            default=schema_.default if not required else ...,
            title=schema_.title,
            description=schema_.description,
            min_length=schema_.minLength,
            max_length=schema_.maxLength,
            pattern=schema_.pattern,
            max_digits=int(schema_.maximum) if schema_.maximum is not None else None,
            lt=int(schema_.exclusiveMaximum)
            if schema_.exclusiveMaximum is not None
            else None,
            lte=int(schema_.maximum) if schema_.maximum is not None else None,
            gt=int(schema_.exclusiveMinimum)
            if schema_.exclusiveMinimum is not None
            else None,
            gte=int(schema_.minimum) if schema_.minimum is not None else None,
        )

        logger.debug(f"field from schema: {field}")

        return field

    def _get_component_schema_by_ref(
        self, *, ref: str
    ) -> Tuple[str | None, Any | None]:
        """Get component from reference.

        Args:
            ref (str): Reference.

        Returns:
            Tuple[str | None, Any | None]: Component name and component object.
        """
        # Handle different types of references
        if ref.startswith("#/components/schemas/"):
            ref = ref.replace("#/components/schemas/", "")
            if ref in self.components.schemas:
                return ref, self.components.schemas[ref]
        elif ref.startswith("#/components/requestBodies/"):
            ref = ref.replace("#/components/requestBodies/", "")
            if hasattr(self.components, 'requestBodies') and ref in self.components.requestBodies:
                return ref, self.components.requestBodies[ref]
        elif ref.startswith("#/components/responses/"):
            ref = ref.replace("#/components/responses/", "")
            if hasattr(self.components, 'responses') and ref in self.components.responses:
                return ref, self.components.responses[ref]
        elif ref.startswith("#/components/parameters/"):
            ref = ref.replace("#/components/parameters/", "")
            if hasattr(self.components, 'parameters') and ref in self.components.parameters:
                return ref, self.components.parameters[ref]
        
        logger.warning(f"Component not found: {ref}")
        return None, None

    def _generate_model_from_component_schema(
        self,
        *,
        class_name: str,
        schema_: Schema,
    ) -> Type[Enum] | Type[BaseModel] | None:
        """Generate model from component schema.

        Args:
            schema_ (Schema): Component schema.

        Raises:
            ValueError: Unsupported schema type.

        Returns:
            Type[BaseModel]: Generated model.
        """
        type_ = schema_.type

        if type_ == SchemaType.STRING.value:
            # for enum
            enum_: List[str] = schema_.enum
            if len(enum_) > 0:
                return Enum(schema_.title, {v: v for v in enum_})
            else:
                raise ValueError(f"Enum is not found: {schema_.title}")
        elif type_ == SchemaType.OBJECT.value:
            # for object
            if schema_.properties:
                return self._generate_model_from_properties(
                    class_name=class_name,
                    required_fields=schema_.required,
                    properties=schema_.properties,
                )
            elif schema_.allOf:
                models: list[Type[Enum] | Type[BaseModel] | None] = []
                for item in schema_.allOf:
                    if isinstance(item, Schema):
                        if item.ref:
                            ref_class_name, component_schema = (
                                self._get_component_schema_by_ref(ref=item.ref)
                            )
                            if isinstance(component_schema, Schema):
                                model_ = self._generate_model_from_component_schema(
                                    class_name=ref_class_name, schema_=component_schema
                                )
                            else:
                                logger.warning(f"Component schema is not a Schema type: {ref_class_name}")
                                model_ = None
                        else:
                            model_ = self._generate_model_from_component_schema(
                                class_name=f"{class_name}_TemporaryClass",  # It's a temporary class name
                                schema_=item,
                            )

                    models.append(model_)

                if len(models) == 0:
                    raise ValueError(f"AllOf items is not found: {schema_.allOf}")
                else:
                    merged_fields = {}
                    for model in models:
                        if model:
                            for field_name, field_info in model.model_fields.items():
                                # @TODO: enhance field definition
                                merged_fields[field_name] = Annotated[
                                    field_info.annotation, field_info
                                ]

                    log_str = (
                        f"Model for request body: {class_name} will be created\n"
                        f"Fields: {len(merged_fields)}\n"
                    )
                    for field_name, field_info in merged_fields.items():
                        log_str += f"  - {field_name}: {field_info}\n"

                    logger.info(log_str)

                    return create_model(
                        f"{class_name}",
                        **merged_fields,
                        __config__=ConfigDict(use_enum_values=True),
                    )
            else:
                raise ValueError(f"Properties is not found: {schema_}")
        else:
            raise ValueError(f"Unsupported schema type: {type_}")

    def _generate_model_from_properties(
        self,
        *,
        class_name: str,
        required_fields: List[str],
        properties: Dict[str, Tuple[Type[Any], Field]],
    ) -> Type[BaseModel]:
        """Generate model from properties.

        Args:
            properties (Dict[str, Tuple[Type[Any], Field]]): Properties.

        Returns:
            Type[BaseModel]: Generated model.
        """
        logger.info(f"class_name: {class_name}")
        logger.info(f"required_fields: {required_fields}")
        # logger.debug(f"properties: {properties}")

        fields = {}
        for field_name, field_schema in properties.items():
            if isinstance(field_schema, Schema):
                field_type = self._get_field_type_from_schema(schema_=field_schema)

                required = (
                    True if required_fields and field_name in required_fields else False
                )

                field = self._make_field_from_schema(
                    schema_=field_schema, required=required
                )

                logger.debug(f"field_type: {field_type}")
                logger.debug(f"field: {field}")

                # @TODO: enhance field definition
                fields[field_name] = Annotated[field_type, field]
            else:
                raise ValueError(f"Unsupported schema type: {type(field_schema)}")

        log_str = (
            f"Model for request body: {class_name} will be created\n"
            f"Fields: {len(fields)}\n"
        )
        for field_name, field_info in fields.items():
            log_str += f"  - {field_name}: {field_info}\n"

        logger.info(log_str)

        return create_model(
            f"{class_name}", **fields, __config__=ConfigDict(use_enum_values=True)
        )



"""This module provides a model for the API operation."""

from __future__ import annotations

import logging
from typing import Type

from pydantic import BaseModel, ConfigDict, create_model

logger = logging.getLogger(__name__)


class AxmpAPIOperation(BaseModel):
    """A model for the API operation."""

    server_name: str
    name: str
    description: str
    path: str
    method: str
    path_params: Type[BaseModel] | None = None
    query_params: Type[BaseModel] | None = None
    request_body: Type[BaseModel] | None = None

    @property
    def args_schema(self) -> Type[BaseModel]:
        """Create the arguments schema for the API operation."""
        model_name = "".join(word.capitalize() for word in self.name.split("_"))
        return self._create_args_schema(
            model_name=f"{model_name}ArgsSchema",
            models=[self.path_params, self.query_params, self.request_body],
        )

    def _create_args_schema(
        self,
        *,
        model_name: str,
        models: list[Type[BaseModel] | None],
    ) -> Type[BaseModel]:
        merged_fields = {}
        for model in models:
            if model:
                for field_name, field_info in model.model_fields.items():
                    merged_fields[field_name] = (field_info.annotation, field_info)

        return create_model(
            model_name, **merged_fields, __config__=ConfigDict(use_enum_values=True)
        )



"""This module contains the models for the OpenAPI specification.

This is from the fastapi.openapi.models.py
Thanks to FastAPI team for the great work!
"""

from enum import Enum
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import AnyUrl, BaseModel, Field
from pydantic.version import VERSION as PYDANTIC_VERSION
from typing_extensions import Annotated, Literal, TypedDict
from typing_extensions import deprecated as typing_deprecated

PYDANTIC_VERSION_MINOR_TUPLE = tuple(int(x) for x in PYDANTIC_VERSION.split(".")[:2])
PYDANTIC_V2 = PYDANTIC_VERSION_MINOR_TUPLE[0] == 2


class BaseModelWithConfig(BaseModel):
    if PYDANTIC_V2:
        model_config = {"extra": "allow"}
    else:

        class Config:
            extra = "allow"


class Contact(BaseModelWithConfig):
    name: Optional[str] = None
    url: Optional[AnyUrl] = None
    email: Optional[str] = None  # removed EmailStr


class License(BaseModelWithConfig):
    name: str
    identifier: Optional[str] = None
    url: Optional[AnyUrl] = None


class Info(BaseModelWithConfig):
    title: str
    summary: Optional[str] = None
    description: Optional[str] = None
    termsOfService: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None
    version: str


class ServerVariable(BaseModelWithConfig):
    enum: Annotated[Optional[List[str]], Field(min_length=1)] = None
    default: str
    description: Optional[str] = None


class Server(BaseModelWithConfig):
    url: Union[AnyUrl, str]
    description: Optional[str] = None
    variables: Optional[Dict[str, ServerVariable]] = None


class Reference(BaseModel):
    ref: str = Field(alias="$ref")


class Discriminator(BaseModel):
    propertyName: str
    mapping: Optional[Dict[str, str]] = None


class XML(BaseModelWithConfig):
    name: Optional[str] = None
    namespace: Optional[str] = None
    prefix: Optional[str] = None
    attribute: Optional[bool] = None
    wrapped: Optional[bool] = None


class ExternalDocumentation(BaseModelWithConfig):
    description: Optional[str] = None
    url: AnyUrl


class Schema(BaseModelWithConfig):
    # Ref: JSON Schema 2020-12: https://json-schema.org/draft/2020-12/json-schema-core.html#name-the-json-schema-core-vocabu
    # Core Vocabulary
    schema_: Optional[str] = Field(default=None, alias="$schema")
    vocabulary: Optional[str] = Field(default=None, alias="$vocabulary")
    id: Optional[str] = Field(default=None, alias="$id")
    anchor: Optional[str] = Field(default=None, alias="$anchor")
    dynamicAnchor: Optional[str] = Field(default=None, alias="$dynamicAnchor")
    ref: Optional[str] = Field(default=None, alias="$ref")
    dynamicRef: Optional[str] = Field(default=None, alias="$dynamicRef")
    defs: Optional[Dict[str, "SchemaOrBool"]] = Field(default=None, alias="$defs")
    comment: Optional[str] = Field(default=None, alias="$comment")
    # Ref: JSON Schema 2020-12: https://json-schema.org/draft/2020-12/json-schema-core.html#name-a-vocabulary-for-applying-s
    # A Vocabulary for Applying Subschemas
    allOf: Optional[List["SchemaOrBool"]] = None
    anyOf: Optional[List["SchemaOrBool"]] = None
    oneOf: Optional[List["SchemaOrBool"]] = None
    not_: Optional["SchemaOrBool"] = Field(default=None, alias="not")
    if_: Optional["SchemaOrBool"] = Field(default=None, alias="if")
    then: Optional["SchemaOrBool"] = None
    else_: Optional["SchemaOrBool"] = Field(default=None, alias="else")
    dependentSchemas: Optional[Dict[str, "SchemaOrBool"]] = None
    prefixItems: Optional[List["SchemaOrBool"]] = None
    # TODO: uncomment and remove below when deprecating Pydantic v1
    # It generales a list of schemas for tuples, before prefixItems was available
    # items: Optional["SchemaOrBool"] = None
    items: Optional[Union["SchemaOrBool", List["SchemaOrBool"]]] = None
    contains: Optional["SchemaOrBool"] = None
    properties: Optional[Dict[str, "SchemaOrBool"]] = None
    patternProperties: Optional[Dict[str, "SchemaOrBool"]] = None
    additionalProperties: Optional["SchemaOrBool"] = None
    propertyNames: Optional["SchemaOrBool"] = None
    unevaluatedItems: Optional["SchemaOrBool"] = None
    unevaluatedProperties: Optional["SchemaOrBool"] = None
    # Ref: JSON Schema Validation 2020-12: https://json-schema.org/draft/2020-12/json-schema-validation.html#name-a-vocabulary-for-structural
    # A Vocabulary for Structural Validation
    type: Optional[str] = None
    enum: Optional[List[Any]] = None
    const: Optional[Any] = None
    multipleOf: Optional[float] = Field(default=None, gt=0)
    maximum: Optional[float] = None
    exclusiveMaximum: Optional[float] = None
    minimum: Optional[float] = None
    exclusiveMinimum: Optional[float] = None
    maxLength: Optional[int] = Field(default=None, ge=0)
    minLength: Optional[int] = Field(default=None, ge=0)
    pattern: Optional[str] = None
    maxItems: Optional[int] = Field(default=None, ge=0)
    minItems: Optional[int] = Field(default=None, ge=0)
    uniqueItems: Optional[bool] = None
    maxContains: Optional[int] = Field(default=None, ge=0)
    minContains: Optional[int] = Field(default=None, ge=0)
    maxProperties: Optional[int] = Field(default=None, ge=0)
    minProperties: Optional[int] = Field(default=None, ge=0)
    required: Optional[List[str]] = None
    dependentRequired: Optional[Dict[str, Set[str]]] = None
    # Ref: JSON Schema Validation 2020-12: https://json-schema.org/draft/2020-12/json-schema-validation.html#name-vocabularies-for-semantic-c
    # Vocabularies for Semantic Content With "format"
    format: Optional[str] = None
    # Ref: JSON Schema Validation 2020-12: https://json-schema.org/draft/2020-12/json-schema-validation.html#name-a-vocabulary-for-the-conten
    # A Vocabulary for the Contents of String-Encoded Data
    contentEncoding: Optional[str] = None
    contentMediaType: Optional[str] = None
    contentSchema: Optional["SchemaOrBool"] = None
    # Ref: JSON Schema Validation 2020-12: https://json-schema.org/draft/2020-12/json-schema-validation.html#name-a-vocabulary-for-basic-meta
    # A Vocabulary for Basic Meta-Data Annotations
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Any] = None
    deprecated: Optional[bool] = None
    readOnly: Optional[bool] = None
    writeOnly: Optional[bool] = None
    examples: Optional[List[Any]] = None
    # Ref: OpenAPI 3.1.0: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#schema-object
    # Schema Object
    discriminator: Optional[Discriminator] = None
    xml: Optional[XML] = None
    externalDocs: Optional[ExternalDocumentation] = None
    example: Annotated[
        Optional[Any],
        typing_deprecated(
            "Deprecated in OpenAPI 3.1.0 that now uses JSON Schema 2020-12, "
            "although still supported. Use examples instead."
        ),
    ] = None


# Ref: https://json-schema.org/draft/2020-12/json-schema-core.html#name-json-schema-documents
# A JSON Schema MUST be an object or a boolean.
SchemaOrBool = Union[Schema, bool]


class Example(TypedDict, total=False):
    summary: Optional[str]
    description: Optional[str]
    value: Optional[Any]
    externalValue: Optional[AnyUrl]

    if PYDANTIC_V2:  # type: ignore [misc]
        __pydantic_config__ = {"extra": "allow"}

    else:

        class Config:
            extra = "allow"


class ParameterInType(Enum):
    query = "query"
    header = "header"
    path = "path"
    cookie = "cookie"


class Encoding(BaseModelWithConfig):
    contentType: Optional[str] = None
    headers: Optional[Dict[str, Union["Header", Reference]]] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None


class MediaType(BaseModelWithConfig):
    schema_: Optional[Union[Schema, Reference]] = Field(default=None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    encoding: Optional[Dict[str, Encoding]] = None


class ParameterBase(BaseModelWithConfig):
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    # Serialization rules for simple scenarios
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None
    schema_: Optional[Union[Schema, Reference]] = Field(default=None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    # Serialization rules for more complex scenarios
    content: Optional[Dict[str, MediaType]] = None


class Parameter(ParameterBase):
    name: str
    in_: ParameterInType = Field(alias="in")


class Header(ParameterBase):
    pass


class RequestBody(BaseModelWithConfig):
    description: Optional[str] = None
    content: Dict[str, MediaType]
    required: Optional[bool] = None


class Link(BaseModelWithConfig):
    operationRef: Optional[str] = None
    operationId: Optional[str] = None
    parameters: Optional[Dict[str, Union[Any, str]]] = None
    requestBody: Optional[Union[Any, str]] = None
    description: Optional[str] = None
    server: Optional[Server] = None


class Response(BaseModelWithConfig):
    description: str
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    content: Optional[Dict[str, MediaType]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None


class Operation(BaseModelWithConfig):
    tags: Optional[List[str]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None
    operationId: Optional[str] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None
    requestBody: Optional[Union[RequestBody, Reference]] = None
    # Using Any for Specification Extensions
    responses: Optional[Dict[str, Union[Response, Any]]] = None
    callbacks: Optional[Dict[str, Union[Dict[str, "PathItem"], Reference]]] = None
    deprecated: Optional[bool] = None
    security: Optional[List[Dict[str, List[str]]]] = None
    servers: Optional[List[Server]] = None


class PathItem(BaseModelWithConfig):
    ref: Optional[str] = Field(default=None, alias="$ref")
    summary: Optional[str] = None
    description: Optional[str] = None
    get: Optional[Operation] = None
    put: Optional[Operation] = None
    post: Optional[Operation] = None
    delete: Optional[Operation] = None
    options: Optional[Operation] = None
    head: Optional[Operation] = None
    patch: Optional[Operation] = None
    trace: Optional[Operation] = None
    servers: Optional[List[Server]] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None


class SecuritySchemeType(Enum):
    apiKey = "apiKey"
    http = "http"
    oauth2 = "oauth2"
    openIdConnect = "openIdConnect"


class SecurityBase(BaseModelWithConfig):
    type_: SecuritySchemeType = Field(alias="type")
    description: Optional[str] = None


class APIKeyIn(Enum):
    query = "query"
    header = "header"
    cookie = "cookie"


class APIKey(SecurityBase):
    type_: SecuritySchemeType = Field(default=SecuritySchemeType.apiKey, alias="type")
    in_: APIKeyIn = Field(alias="in")
    name: str


class HTTPBase(SecurityBase):
    type_: SecuritySchemeType = Field(default=SecuritySchemeType.http, alias="type")
    scheme: str


class HTTPBearer(HTTPBase):
    scheme: Literal["bearer"] = "bearer"
    bearerFormat: Optional[str] = None


class OAuthFlow(BaseModelWithConfig):
    refreshUrl: Optional[str] = None
    scopes: Dict[str, str] = {}


class OAuthFlowImplicit(OAuthFlow):
    authorizationUrl: str


class OAuthFlowPassword(OAuthFlow):
    tokenUrl: str


class OAuthFlowClientCredentials(OAuthFlow):
    tokenUrl: str


class OAuthFlowAuthorizationCode(OAuthFlow):
    authorizationUrl: str
    tokenUrl: str


class OAuthFlows(BaseModelWithConfig):
    implicit: Optional[OAuthFlowImplicit] = None
    password: Optional[OAuthFlowPassword] = None
    clientCredentials: Optional[OAuthFlowClientCredentials] = None
    authorizationCode: Optional[OAuthFlowAuthorizationCode] = None


class OAuth2(SecurityBase):
    type_: SecuritySchemeType = Field(default=SecuritySchemeType.oauth2, alias="type")
    flows: OAuthFlows


class OpenIdConnect(SecurityBase):
    type_: SecuritySchemeType = Field(
        default=SecuritySchemeType.openIdConnect, alias="type"
    )
    openIdConnectUrl: str


SecurityScheme = Union[APIKey, HTTPBase, OAuth2, OpenIdConnect, HTTPBearer]


class Components(BaseModelWithConfig):
    schemas: Optional[Dict[str, Union[Schema, Reference]]] = None
    responses: Optional[Dict[str, Union[Response, Reference]]] = None
    parameters: Optional[Dict[str, Union[Parameter, Reference]]] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    requestBodies: Optional[Dict[str, Union[RequestBody, Reference]]] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    securitySchemes: Optional[Dict[str, Union[SecurityScheme, Reference]]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None
    # Using Any for Specification Extensions
    callbacks: Optional[Dict[str, Union[Dict[str, PathItem], Reference, Any]]] = None
    pathItems: Optional[Dict[str, Union[PathItem, Reference]]] = None


class Tag(BaseModelWithConfig):
    name: str
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None


class OpenAPI(BaseModelWithConfig):
    openapi: str
    info: Info
    jsonSchemaDialect: Optional[str] = None
    servers: Optional[List[Server]] = None
    # Using Any for Specification Extensions
    paths: Optional[Dict[str, Union[PathItem, Any]]] = None
    webhooks: Optional[Dict[str, Union[PathItem, Reference]]] = None
    components: Optional[Components] = None
    security: Optional[List[Dict[str, List[str]]]] = None
    tags: Optional[List[Tag]] = None
    externalDocs: Optional[ExternalDocumentation] = None
